/*
A GLUT program using OpenGL, GLUI, and GLSL.

Demonstrates global illumination--computing shadowed form factors.

Orion Sky Lawlor, olawlor@acm.org, 2007/04/12 (Public Domain)
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <GL/glew.h> /* GL Extensions Wrangler, http://glew.sourceforge.net/ */
#include <GL/glut.h> /* GL Utilities Toolkit, http://freeglut.sourceforge.net/ */
#include <GL/glui.h> /* GL User Interface, http://glui.sourceforge.net/ */
#include "osl/vec4.h" /* GLSL-style C++ types, by Orion Lawlor */
#include "osl/mat4.h"
using osl::mat4;

/* Dr. Lawlor's GLSL utilities */
#include "ogl/program.h"
#include "ogl/program.cpp"
#include "ogl/util.cpp"
#include "ogl/fast_mipmaps.c"

#include "ogl/minicam.h"


/* Variables updated by the GUI */
int win_w, win_h; /**< Width and height (in pixels) of our window at last resize */

float brightness=2.0, lightsize=1.0, spheresize=0.5, shininess=1.0;

/* These are rotation matrices, updated by the GLUI_Rotation spheres */
mat4 lightMat(1.0);

/// Scale by this offset
inline void glScalefv(const float *s) {
	glScalef(s[0],s[1],s[2]);
}
/// Translate by this offset
inline void glTranslatefv(const float *s) {
	glTranslatef(s[0],s[1],s[2]);
}
/** Transpose this mat4.  If the mat4 was a rotation matrix, 
  this actually inverts the matrix. */
mat4 transpose(const mat4 &m) {
	return mat4(
		m[0][0], m[1][0], m[2][0], m[3][0],
		m[0][1], m[1][1], m[2][1], m[3][1],
		m[0][2], m[1][2], m[2][2], m[3][2],
		m[0][3], m[1][3], m[2][3], m[3][3]);
}


/* Draw a big green ground polygon */
void draw_ground(void) {
	double ground=100.0; /* extent of ground in X and Y */
	double groundZ=-0.8;
	glColor4f(0.1,0.4,0.1,1); /* green */
	glNormal3f(0,0,1); /* facing straight up */
	oglTextureQuad(
		oglVector3d(-ground,-ground,groundZ), oglVector3d(+ground,-ground,groundZ), 
		oglVector3d(-ground,+ground,groundZ), oglVector3d(+ground,+ground,groundZ)
	);
}

// These are the corners of our light source.
//  They get transformed by lightMat and lightsize.
vec3 lightCorners[3]={vec3(-1,0,0), vec3(-1,-1,0), vec3(-1,0,1) };

/* GLUT calls this routine every time our window needs to be redrawn */
void main_display(void) {
	win_w=glutGet(GLUT_WINDOW_WIDTH); win_h=glutGet(GLUT_WINDOW_HEIGHT);
	glViewport(0, 0, (GLsizei)win_w, (GLsizei)(win_h));
	
	// Clear the color and depth buffers
	//glClearColor(0.2,0.4,0.8,0.0); /* sky blue */
	float g=0.05; glClearColor(g,g,g,0.0); /* gray */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); /* actually clear color and depth */
	
	/* Depth testing */
	glEnable(GL_DEPTH_TEST);
	
	/* Alpha blending */
	glEnable(GL_BLEND);
	           /*  src color  + dest color  */
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	// Load all needed matrices into OpenGL
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); /*<- clean out any old leftover matrices */
	gluPerspective(90.0, /* <- Y camera field-of-view, in degrees */
		win_w/(float)win_h, /* viewport's aspect ratio: always width over height */
		0.01, /* Near clipping plane depth */
		100.0 /* Far clipping plane depth */
	);
	oglCameraLookAt(0.1+length(camera));
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
/* Set up programmable shader for lit objects */
	oglProgramObject *p=oglProgramFromFiles("vertex.txt","fragment.txt");
	p->begin();
	p->set("brightness",brightness); /* full diffuse intensity */
	p->set("cameraloc",camera);
	if (0) { /* show the camera position */
		glPushMatrix();
		glTranslatefv(camera);
		glutSolidSphere(0.4,20,30);
		glPopMatrix();
	}
	p->set("shininess",1.0); /* full specular intensity */
	p->set("lightCorner0",lightMat*lightsize*vec4(lightCorners[0],1.0));
	p->set("lightCorner1",lightMat*lightsize*vec4(lightCorners[1],1.0));
	p->set("lightCorner2",lightMat*lightsize*vec4(lightCorners[2],1.0));
	p->set("spheresize",spheresize);
	p->set("blockSphere",vec4(0,0,0,0.5));
	p->set("screenSize",vec4(win_w,win_h,1,1));
	
	/* Draw a big green ground polygon */
	draw_ground();
	
	// Draw the world geometry
	glColor4f(1,1,0,1); /* yellow */
	//glutSolidTorus(0.2,0.5,40,60);
	glColor4f(0.5,0.5,0.5,1); /* gray */
	glutSolidSphere(spheresize,40,30);
	
	p->end();
	delete p;
	
// Draw the light source polygon
	glScalef(lightsize,lightsize,lightsize);
	glMultMatrixf(&lightMat[0][0]);
	glColor3f(1,1,1);
	glBegin(GL_TRIANGLE_FAN);
	for (int i=0;i<3;i++) glVertex3fv(lightCorners[i]);
	glEnd();
	
	// Display the stuff drawn so far to the screen:
	glutSwapBuffers();
}

int main(int argc, char** argv)
{
	glutInitWindowSize(800,600);
	glutInit(&argc,argv);

	// Ask for a back buffer, color buffer, and Z buffer in a window.
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
	glutInitDisplayMode (mode);
	int main_window = glutCreateWindow("GLSL Lighting Demo");

	// Turn on OpenGL extensions, like programmable shaders
	//   WARNING: you MUST call glutCreateWindow before doing this!
	//   If you don't call glewInit, or call it wrong, you'll segfault!
	glewInit();
	
	// Ask GLUT to call us when we need to redraw
	glutDisplayFunc (main_display);
	
	oglCameraInit();
	camera=vec3(0.0,-2.0,0.0);

	if (1) {
		// Make some sliders (this is the only GLUI code you'll need!)
		GLUI *glui = GLUI_Master.create_glui_subwindow(main_window, 
			GLUI_SUBWINDOW_TOP);
		/* This causes the main window to be redrawn when a GLUI control changes */
		glui->set_main_gfx_window( main_window );

		(new GLUI_Spinner( glui, "Lightsize:", &lightsize))
			->set_float_limits(0.0, 10.0);
		(new GLUI_Spinner( glui, "Spheresize:", &spheresize))
			->set_float_limits(0.0, 10.0);
		(new GLUI_Spinner( glui, "Shutter Speed:", &brightness))
			->set_float_limits(0.0, 100.0);
		
		new GLUI_Column(glui);
		new GLUI_Rotation(glui,"Light:",&lightMat[0][0]);
	}
	
	// Process events until user closes our window...
	glutMainLoop();
	
	return 0;
}
